package lsedit;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import java.util.Enumeration;

import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;

public class QueryBox extends JComponent implements MouseListener
{
	protected static final String g_duplicateRelations		   = "Duplicate Relations";
	protected static final String g_liftActiveRelations		   = "Lift Active Relations";
	protected static final String g_deleteActiveEntities       = "Delete Active Entities";
	protected static final String g_deleteActiveContainers     = "Delete Active Containers";
	protected static final String g_deleteActiveRelations      = "Delete Active Relations";
	protected static final String g_resetIOpoints			   = "Reset IO points";

	protected static final String g_groupWithQueries = "Group with queries";

	private class ActiveEntityChkBox extends EntityChkBox implements ItemListener
	{
		public ActiveEntityChkBox(EntityClass ec, int index, int h) 
		{
			super(ec, index, -1, h, false);

			setFont(m_textFont);
			addItemListener(this);
		}

		public boolean isActive()
		{
			return(m_ec.isActive());
		}

		public void setActive(boolean value)
		{
			if (m_ec.isActive() != value) {
				m_ec.setActive(value);
				repaint();
		}	}

		public void itemStateChanged(ItemEvent ev)
		{
//			System.out.println("EntityLabel.itemStateChanged " + ev.getStateChange() + " " + isSelected());
			setActive(ev.getStateChange() == ItemEvent.SELECTED);
	}	}

	class ActiveRelnChkBox extends RelnChkBox /* extends JComponent */ implements ItemListener, MouseListener
	{
		protected	JPopupMenu			m_popup;

		public ActiveRelnChkBox(RelationClass rc,  int index, int count, Font font) 
		{
			super(rc, index, count, rc.isActive(), font);
			addItemListener(this);
			addMouseListener(this);
		}

		public boolean isActive()
		{
			return(m_rc.isActive());
		}

		public void setActive(boolean value)
		{
			if (m_rc.isActive() != value) {
				m_rc.setActive(value);
				repaint();
		}	}

		public void itemStateChanged(ItemEvent ev)
		{
			setActive(ev.getStateChange() == ItemEvent.SELECTED);
		}

		// MouseListener interface

		public void mouseClicked(MouseEvent e)
		{
		}

		public void mouseEntered(MouseEvent e)
		{
		}

		public void mouseExited(MouseEvent e)
		{
		}

		public void mousePressed(MouseEvent ev)
		{
			if (ev.isMetaDown()) {

				int		x           = ev.getX();
				int		y           = ev.getY();

				JMenuItem	mi;
				Diagram		diagram = m_ls.getDiagram();

				m_popup = new JPopupMenu("Relation menu");

				mi = new JMenuItem(g_duplicateRelations);
				mi.addActionListener(new DuplicateEdges(m_rc));
				m_popup.add(mi);

				FontCache.setMenuTreeFont(m_popup); 
				QueryBox.this.add(m_popup);
//				Do.dump_menu(m_popup);
				m_popup.show(this, x, y);
		}	}
		
		public void mouseReleased(MouseEvent ev) 
		{
		}
	}

	protected class GrpChkBox extends JCheckBox implements ItemListener
	{
		GrpChkBox(String label)
		{
			super(label);

			setSelected(m_ls.isGroupQuery());
			addItemListener(this);
		}

		protected void grpStateInfo() 
		{
			m_ls.doFeedback("Entities " + (m_ls.isGroupQuery() ? "are" : "aren't") + " grouped with queries");
		}

		public void itemStateChanged(ItemEvent ev)
		{
			m_ls.setGroupQuery(isSelected()); 
			grpStateInfo();
			m_ls.requestFocus();
		}
	}	

	public final static String DEFAULT_QUERY_TITLE_FONT_NAME  = FontCache.DEFAULT_FONT_NAME;
	public final static int    DEFAULT_QUERY_TITLE_FONT_STYLE = Font.BOLD;
	public final static int    DEFAULT_QUERY_TITLE_FONT_SIZE  = 12;

	protected static Font m_titleFont   = FontCache.get(DEFAULT_QUERY_TITLE_FONT_NAME, DEFAULT_QUERY_TITLE_FONT_STYLE, DEFAULT_QUERY_TITLE_FONT_SIZE);

	public final static String DEFAULT_QUERY_TEXT_FONT_NAME  = FontCache.DEFAULT_FONT_NAME;
	public final static int    DEFAULT_QUERY_TEXT_FONT_STYLE = Font.PLAIN;
	public final static int    DEFAULT_QUERY_TEXT_FONT_SIZE  = 11;

	protected static Font m_textFont    = FontCache.get(DEFAULT_QUERY_TEXT_FONT_NAME, DEFAULT_QUERY_TEXT_FONT_STYLE, DEFAULT_QUERY_TEXT_FONT_SIZE);

	public    static final String m_helpStr = "This box shows the active entities and relations during update and queries";

	protected static final int MARGIN = 5;
	protected static final int GAP    = 5; 

	protected LandscapeEditorCore   m_ls;
	protected JTabbedPane			m_tabbedPane;
	protected JScrollPane			m_scrollPane;

	// Objects used in layout

	protected JLabel				m_ulabel;
	protected JLabel				m_elabel2;
	protected JLabel				m_rlabel;
	protected JLabel				m_clabel;
	protected JLabel				m_footer1;
	protected JLabel				m_footer2;

	protected int					m_width;
	protected int					m_height;

	protected GrpChkBox				m_grpChk;

	// --------------
	// Public methods 
	// --------------

	public QueryBox(LandscapeEditorCore ls, JTabbedPane tabbedPane) 
	{
		m_ls         = ls;
		m_tabbedPane = tabbedPane;

		setBackground(Diagram.boxColor);

		m_ulabel = new JLabel("Active Entities");
		m_ulabel.setForeground(Color.red);
		m_ulabel.setFont(m_titleFont);

		m_elabel2 = new JLabel("Right click for menu.");
		m_elabel2.setForeground(Color.black);
		m_elabel2.setFont(m_textFont);

		m_rlabel = new JLabel("Active Relations");
		m_rlabel.setForeground(Color.red);
		m_rlabel.setFont(m_titleFont);

		m_clabel = new JLabel("Hierarchy");
		m_clabel.setForeground(Color.red);
		m_clabel.setFont(m_titleFont);

		m_footer1 = new JLabel("Checkboxes select query relations.");
		m_footer1.setForeground(Color.black);
		m_footer1.setFont(m_textFont);

		m_footer2 = new JLabel("Click relation arrow for description.");
		m_footer2.setForeground(Color.black);
		m_footer2.setFont(m_textFont);

		m_grpChk = new GrpChkBox(g_groupWithQueries);
		m_grpChk.setForeground(Color.black);
		m_grpChk.setFont(m_textFont);

		setLayout(null);
		setBackground(Diagram.boxColor);
		setToolTipText(m_helpStr);

		m_scrollPane = new JScrollPane();
		m_scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED); 
		m_scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED); 

		m_scrollPane.setViewportView(this);

		tabbedPane.addTab("Query", null, m_scrollPane, m_helpStr);
		addMouseListener(this);
	}

	public boolean isActive() 
	{
		Component active;

		if (!isVisible()) {
			return(false);
		}
		active = m_tabbedPane.getSelectedComponent();
		return(active != null && m_scrollPane == active);
	}	

	public void activate() 
	{
		m_tabbedPane.setSelectedComponent(m_scrollPane);
	}

	public void setGroupQuery(boolean value)
	{
		if (m_grpChk.isSelected() != value) {
			m_grpChk.setSelected(value);
	}	}

	public void containsClassChanged()
	{
		fillQueryBox();
	}

	public static Font getTitleFont()
	{
		return m_titleFont;
	}

	public static void setTitleFont(Font font)
	{
		m_titleFont = font;
	}

	public void titleFontChanged(Font font)
	{
		m_ulabel.setFont(font);
		fillQueryBox();
	}

	public static Font getTextFont()
	{
		return m_textFont;
	}

	public static void setTextFont(Font font)
	{
		m_textFont = font;
	}

	public void textFontChanged(Font font)
	{
		m_footer1.setFont(font);
		m_footer2.setFont(font);
		m_grpChk.setFont(font);
		fillQueryBox();
	}

	protected void add(JComponent component)
	{
		Dimension	d;
		int			width, height;

		d         = component.getPreferredSize();
		width     = d.width;
		height    = d.height;
		if (width > m_width) {
//			System.out.println("Width=" + width + " for " + component);
			m_width = width;
		}
		component.setBounds(MARGIN, m_height, width, height);
		super.add(component);
		m_height += height;
	}

	public void fillQueryBox() 
	{
		Diagram				diagram;
		FontMetrics			fontMetrics;
		Enumeration			en;
		int					item,n,h;
		RelationClass		rc, contains;
		Color				color;
		ActiveRelnChkBox	relnChk;
		Dimension			dimension;

		diagram  = m_ls.getDiagram();

		removeAll();

		m_width  = 0;
		m_height = 0;

		// Draw legend for entities

		item = 0;
		n	 = 0;

		fontMetrics = getFontMetrics(m_textFont);
		h           = fontMetrics.getHeight();

		EntityClass ec;
		String		id;

		m_height += 10;
		add(m_ulabel);

		if (diagram != null) {
			for (en = diagram.enumEntityClassesInOrder(); en.hasMoreElements(); ) {
				ec = (EntityClass) en.nextElement();
				m_height += 10;
				add(new ActiveEntityChkBox(ec, ++n, h));
		}	}

		m_height += 10;
		add(m_elabel2);
		m_height += 20;
		add(m_rlabel);

		// Draw legend for relations

		m_height += GAP;

		contains = null;
		if (diagram != null) {
			m_height += GAP;
			for (en = diagram.enumRelationClassesInOrder(); en.hasMoreElements(); ) {

				rc = (RelationClass) en.nextElement();

				if (!rc.isContainsClass()) {
					add(new ActiveRelnChkBox(rc, ++n, -1, m_textFont));
				} else {
					contains = rc;
		}	}	}
		
		if (contains != null) {
			m_height += 10;
			add(m_clabel);
			m_height += GAP;
			add(new ActiveRelnChkBox(contains, ++n, -1, m_textFont));
		}

		m_height += 20;
		add(m_footer1);
		add(m_footer2);
		m_height += 10;
		add(m_grpChk);

		dimension = new Dimension(m_width, m_height);
		setPreferredSize(dimension);
		setBounds(0, 0, m_width, m_height);
	}

	public void toggleRelationActivity(int key) 
	{
		int				cnt = getComponentCount();
		int				i;
		Component		c;
		boolean			ns;
		ActiveRelnChkBox box;

		if (key == 0) {
			ns = true;
			for (i = 0; i < cnt; ++i) {
				c = getComponent(i);
				if (c instanceof ActiveRelnChkBox) {
					box = (ActiveRelnChkBox) c;
					if (box.getIndex() == 1) {
						// Set everything same way as first
						ns = !box.isActive();
						break;
			}	}	}

			for (i = 0; i < cnt; ++i) {
				c = getComponent(i);
				if (c instanceof ActiveRelnChkBox) {
					box = (ActiveRelnChkBox) c;
					if (box.isActive() != ns) {
						if (!box.isContainsClass()) {
							box.doClick();
			}	}	}	}
		} else {
			for (i = 0; i < cnt; ++i) {
				c = getComponent(i);
				if (c instanceof ActiveRelnChkBox) {
					if (((ActiveRelnChkBox) c).getIndex() == key) {
						((ActiveRelnChkBox) c).doClick();
						break;
		}	}	}	}
	}

	private class SetEntitiesActive implements ActionListener
	{
		boolean	m_active;

		public SetEntitiesActive(boolean active)
		{
			m_active = active;
		}

		public void actionPerformed(ActionEvent ev)
		{
			if ((ev.getModifiers() & ActionEvent.META_MASK) != 0) {
				if (m_ls.processMetaKeyEvent("Set Entities Active")) {
					return;
			}	}

			int				cnt    = getComponentCount();
			boolean			active = m_active;
			int				i;
			Component		c;
			ActiveEntityChkBox box;

			for (i = 0; i < cnt; ++i) {
				c = getComponent(i);
				if (c instanceof ActiveEntityChkBox) {
					box = (ActiveEntityChkBox) c;
					if (box.isActive() != active) {
						box.doClick();
		}	}	}	}	
	}

	private class SetRelationsActive implements ActionListener
	{
		boolean	m_active;

		public SetRelationsActive(boolean active)
		{
			m_active = active;
		}

		public void actionPerformed(ActionEvent ev)
		{
			if ((ev.getModifiers() & ActionEvent.META_MASK) != 0) {
				if (m_ls.processMetaKeyEvent("Set Relations Active")) {
					return;
			}	}

			int				cnt    = getComponentCount();
			boolean			active = m_active;
			int				i;
			Component		c;
			ActiveRelnChkBox box;

			for (i = 0; i < cnt; ++i) {
				c = getComponent(i);
				if (c instanceof ActiveRelnChkBox) {
					box = (ActiveRelnChkBox) c;
					if (!box.isContainsClass() && box.isActive() != active) {
						box.doClick();
		}	}	}	}	
	}

	private class LiftEntityClassEdges implements ActionListener
	{
		public LiftEntityClassEdges()
		{
		}

		public void actionPerformed(ActionEvent ev)
		{
			if ((ev.getModifiers() & ActionEvent.META_MASK) != 0) {
				if (m_ls.processMetaKeyEvent(g_liftActiveRelations)) {
					return;
			}	}

			switch (JOptionPane.showConfirmDialog(null, "Raise active edges in active entities", "Lift edges to parent else delete", JOptionPane.YES_NO_CANCEL_OPTION)) {
			case JOptionPane.YES_OPTION:
			{
				Diagram		diagram = m_ls.getDiagram();

				m_ls.beginUndoRedo(g_liftActiveRelations);
				diagram.updateLiftEdges();
				m_ls.endUndoRedo();
				diagram.redrawDiagram();
				break;
			}
			case JOptionPane.NO_OPTION:
				break;
			}
	}	}

	private class DeleteActiveEntities implements ActionListener
	{
		public DeleteActiveEntities()
		{
		}

		public void actionPerformed(ActionEvent ev)
		{
			if ((ev.getModifiers() & ActionEvent.META_MASK) != 0) {
				if (m_ls.processMetaKeyEvent(g_deleteActiveEntities)) {
					return;
			}	}

			switch (JOptionPane.showConfirmDialog(null, "Delete active entities below draw root", "Delete entities", JOptionPane.YES_NO_CANCEL_OPTION)) {
			case JOptionPane.YES_OPTION:
			{
				Diagram		diagram = m_ls.getDiagram();

				m_ls.beginUndoRedo(g_deleteActiveEntities);
				diagram.updateDeleteActiveEntities();
				m_ls.endUndoRedo();
				diagram.redrawDiagram();
				break;
			}
			case JOptionPane.NO_OPTION:
				break;
			}
	}	}

	private class DeleteActiveContainers implements ActionListener
	{
		public DeleteActiveContainers()
		{
		}

		public void actionPerformed(ActionEvent ev)
		{
			if ((ev.getModifiers() & ActionEvent.META_MASK) != 0) {
				if (m_ls.processMetaKeyEvent(g_deleteActiveContainers)) {
					return;
			}	}

			switch (JOptionPane.showConfirmDialog(null, "Delete active containers below draw root", "Delete containers", JOptionPane.YES_NO_CANCEL_OPTION)) {
			case JOptionPane.YES_OPTION:
			{
				Diagram		diagram = m_ls.getDiagram();

				m_ls.beginUndoRedo(g_deleteActiveContainers);
				diagram.updateDeleteActiveContainers();
				m_ls.endUndoRedo();
				diagram.redrawDiagram();
				break;
			}
			case JOptionPane.NO_OPTION:
				break;
			}
	}	}

	private class DuplicateEdges implements ActionListener
	{
		RelationClass m_rc;

		public DuplicateEdges(RelationClass rc)
		{
			m_rc = rc;
		}

		public void actionPerformed(ActionEvent ev)
		{	
			if ((ev.getModifiers() & ActionEvent.META_MASK) != 0) {
				if (m_ls.processMetaKeyEvent(g_duplicateRelations)) {
					return;
			}	}
			m_ls.duplicateEdges(m_rc);
		}
	}

	private class DeleteActiveEdges implements ActionListener
	{
		public DeleteActiveEdges()
		{
		}

		public void actionPerformed(ActionEvent ev)
		{
			if ((ev.getModifiers() & ActionEvent.META_MASK) != 0) {
				if (m_ls.processMetaKeyEvent(g_deleteActiveRelations)) {
					return;
			}	}

			switch (JOptionPane.showConfirmDialog(null, "Delete active relations in active entities below draw root", "Delete edges", JOptionPane.YES_NO_CANCEL_OPTION)) {
			case JOptionPane.YES_OPTION:
			{
				Diagram		diagram = m_ls.getDiagram();

				m_ls.beginUndoRedo(g_deleteActiveRelations);
				diagram.updateDeleteActiveEdges();
				m_ls.endUndoRedo();
				diagram.redrawDiagram();
				break;
			}
			case JOptionPane.NO_OPTION:
				break;
			}
	}	}


	private class ResetIOpoints implements ActionListener
	{
		public ResetIOpoints()
		{
		}

		public void actionPerformed(ActionEvent ev)
		{
			if ((ev.getModifiers() & ActionEvent.META_MASK) != 0) {
				if (m_ls.processMetaKeyEvent(g_resetIOpoints)) {
					return;
			}	}
			Diagram diagram = m_ls.getDiagram();
			if (diagram != null) {
				diagram.resetIOpoints();
				diagram.repaint();
	}	}	}

	// Generic 	MouseListener interface

	protected void doRightPopup(MouseEvent ev)
	{
		Diagram			diagram = m_ls.getDiagram();
		Enumeration		en;
		EntityClass		ec;
		RelationClass	rc;
		int				x, y;
		JPopupMenu		popupMenu;
		JMenuItem		mi;
		String			string;


		if (diagram == null) {
			return;
		}

		x         = ev.getX();
		y         = ev.getY();
		popupMenu = new JPopupMenu("Update options");

		for (en = diagram.enumEntityClasses(); en.hasMoreElements(); ) {
			ec = (EntityClass) en.nextElement();
			if (ec.isActive()) {
				mi = new JMenuItem("No entities active");
				mi.addActionListener(new SetEntitiesActive(false));
				popupMenu.add(mi);
				break;
		}	}

		for (en = diagram.enumEntityClasses(); en.hasMoreElements(); ) {
			ec = (EntityClass) en.nextElement();
			if (!ec.isActive()) {
				mi = new JMenuItem("All entities active");
				mi.addActionListener(new SetEntitiesActive(true));
				popupMenu.add(mi);
				break;
		}	}
	
		for (en = diagram.enumRelationClasses(); en.hasMoreElements(); ) {
			rc = (RelationClass) en.nextElement();
			if (!rc.isContainsClass() && rc.isActive()) {
				mi = new JMenuItem("No relations active");
				mi.addActionListener(new SetRelationsActive(false));
				popupMenu.add(mi);
				break;
		}	}

		for (en = diagram.enumRelationClasses(); en.hasMoreElements(); ) {
			rc = (RelationClass) en.nextElement();
			if (!rc.isContainsClass() && !rc.isActive()) {
				mi = new JMenuItem("All relations active");
				mi.addActionListener(new SetRelationsActive(true));
				popupMenu.add(mi);
				break;
		}	}

		mi = new JMenuItem(g_liftActiveRelations);
		mi.addActionListener(new LiftEntityClassEdges());
		popupMenu.add(mi);

		mi = new JMenuItem(g_deleteActiveEntities);
		mi.addActionListener(new DeleteActiveEntities());
		popupMenu.add(mi);

		mi = new JMenuItem(g_deleteActiveContainers);
		mi.addActionListener(new DeleteActiveContainers());
		popupMenu.add(mi);

		mi = new JMenuItem(g_deleteActiveRelations);
		mi.addActionListener(new DeleteActiveEdges());
		popupMenu.add(mi);

		mi = new JMenuItem(g_resetIOpoints);
		mi.addActionListener(new ResetIOpoints());
		popupMenu.add(mi);

		FontCache.setMenuTreeFont(popupMenu); 
		add(popupMenu);
		popupMenu.show(this, x, y);
//		Do.dump_menu(popupMenu);
		remove(popupMenu);
	}

	public void mouseClicked(MouseEvent ev)
	{
		if (ev.isMetaDown()) {
			doRightPopup(ev);
	}	}

	public void mouseEntered(MouseEvent ev)
	{
	}

	public void mouseExited(MouseEvent e)
	{
	}

	public void mousePressed(MouseEvent ev)
	{
	}

	public void mouseReleased(MouseEvent ev)
	{
	}
}

